/** fMSX: portable MSX emulator ******************************/
/**                                                         **/
/**                           Unix.c                        **/
/**                                                         **/
/** This file contains Unix/X-dependent subroutines and     **/
/** drivers. It includes common drivers from Common.h.      **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1994-1999                 **/
/**               Elan Feingold   1995                      **/
/**               Ville Hallik    1996                      **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/
#ifdef UNIX

/** Private #includes ****************************************/
#include "MSX.h"
#include "LibUnix.h"

/** Standard Unix/X #includes ********************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

/** Public parameters ****************************************/
char *Title="fMSX Unix/X 2.1";  /* Window/command line title */
int SaveCPU  = 1;               /* 1 = freeze when focus out */
int UseSHM   = 1;               /* 1 = use MITSHM            */
int UseZoom  = 1;               /* Window zoom factor        */
int SyncFreq = 0;               /* Screen update frequency   */
char *Disks[2][MAXDISK+1];      /* Disk names for each drive */

/** Various variables ****************************************/
#define WIDTH  324
#define HEIGHT 228

static volatile byte TimerReady;
static unsigned int BPal[256],XPal[16],XPal0; 
static char *XBuf;
static Port Prt;
static byte JoyState;
static int CurDisk[2];

/** Various X-related variables ******************************/
static Display *Dsp;
static Colormap DefaultCMap;
static GC DefaultGC;
static unsigned long White,Black;

/** Sound-related definitions ********************************/
#ifdef SOUND
int InitSound(void);
void TrashSound(void);
void FlushSound(void);
void StopSound(void);
void ResumeSound(void);
extern int UseSound;
#endif

/** These functions are called on signals ********************/
static void OnBreak(int Arg) { ExitNow=1;signal(Arg,OnBreak); }
static void OnTimer(int Arg) { TimerReady=1;signal(Arg,OnTimer); }

/** Keyboard bindings ****************************************/
byte Keys[256][2] =
{
/* 00 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {7,0x20},{7,0x08},{0,0x00},{0,0x00},{0,0x00},{7,0x80},{0,0x00},{0,0x00},
/* 10 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{7,0x04},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* 20 */
  {8,0x01},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{2,0x01},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{2,0x04},{1,0x04},{2,0x08},{2,0x10},
/* 30 */
  {0,0x01},{0,0x02},{0,0x04},{0,0x08},{0,0x10},{0,0x20},{0,0x40},{0,0x80},
  {1,0x01},{1,0x02},{0,0x00},{1,0x80},{0,0x00},{1,0x08},{0,0x00},{0,0x00},
/* 40 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* 50 */
  {0,0x00},{8,0x10},{8,0x20},{8,0x80},{8,0x40},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{1,0x20},{1,0x10},{1,0x40},{0,0x00},{0,0x00},
/* 60 */
  {2,0x02},{2,0x40},{2,0x80},{3,0x01},{3,0x02},{3,0x04},{3,0x08},{3,0x10},
  {3,0x20},{3,0x40},{3,0x80},{4,0x01},{4,0x02},{4,0x04},{4,0x08},{4,0x10},
/* 70 */
  {4,0x20},{4,0x40},{4,0x80},{5,0x01},{5,0x02},{5,0x04},{5,0x08},{5,0x10},
  {5,0x20},{5,0x40},{5,0x80},{0,0x00},{0,0x00},{0,0x00},{6,0x04},{0,0x00},
/* 80 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* 90 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* A0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* B0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{6,0x20},{6,0x40},
/* C0 */
  {6,0x80},{7,0x01},{7,0x02},{8,0x02},{8,0x08},{8,0x04},{7,0x40},{7,0x10},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* D0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {8,0x02},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* E0 */
  {0,0x00},{6,0x01},{6,0x08},{6,0x02},{6,0x10},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{6,0x04},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* F0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{8,0x08}
};

/** InitMachine() ********************************************/
/** Allocate resources needed by Unix/X-dependent code.     **/
/*************************************************************/
int InitMachine(void)
{
  Screen *Scr;
  int J;

  /* Reset all variables */
  XBuf=NULL;JoyState=0x00;
  CurDisk[0]=CurDisk[1]=0;

  /* Correct configuration values */
  UseZoom  = UseZoom>4? 4:UseZoom<1? 1:UseZoom;
  SyncFreq = SyncFreq>60? 60:SyncFreq<0? 0:SyncFreq;

  /* Open display */
  if(Verbose) printf("Initializing Unix/X drivers:\n  Opening display...");
  if(!(Dsp=XOpenDisplay(NULL))) { if(Verbose) puts("FAILED");return(0); }

  /* Set internal variables */
  Scr=DefaultScreenOfDisplay(Dsp);
  White=WhitePixelOfScreen(Scr);
  Black=BlackPixelOfScreen(Scr);
  DefaultGC=DefaultGCOfScreen(Scr);
  DefaultCMap=DefaultColormapOfScreen(Scr);

  /* Reset the palette */
  XPal0=Black;
  for(J=0;J<16;J++) XPal[J]=Black;

  /* Keys do not autorepeat */
  XAutoRepeatOff(Dsp);

  /* Initialize LibUnix toolkit */
  if(Verbose) printf("OK\n  Initializing LibUnix...");
  if(!InitLibUnix(Dsp)) { if(Verbose) puts("FAILED");return(0); }

  /* Create a viewport */
  if(Verbose) printf("OK\n  Opening window...");
  if(!X11NewPort(&Prt,Title,WIDTH,HEIGHT,UseSHM? USE_SHM:0))
  { if(Verbose) puts("FAILED");return(0); }
  if(Verbose) printf("%dbpp...OK\n",Prt.Img.D);
  XBuf=Prt.Img.Data;
  
  /* Set image/window size */
  X11SetPort(&Prt,0,0,WIDTH,HEIGHT);
  X11ZoomPort(&Prt,UseZoom*WIDTH,UseZoom*HEIGHT);

  /* Allocate colors for SCREEN8 */
  {
    XColor Color;
    word R,G,B;
    byte I;

    for(R=0;R<4;R++)
      for(G=0;G<4;G++)
        for(B=0;B<4;B++)
        {
          Color.flags=DoRed|DoGreen|DoBlue;
          Color.red=R<<14;
          Color.green=G<<14;
          Color.blue=B<<14;
          I=B|(R<<3)|(G<<6);

          BPal[I]      =
          BPal[I|0x04] =
          BPal[I|0x20] =
          BPal[I|0x24] =
            XAllocColor(Dsp,DefaultCMap,&Color)? Color.pixel:Black;
        }  
  }

#ifdef SOUND
  /* Initialize sound */
  if(UseSound) InitSound();
#endif SOUND

  /* Catch all signals */
  signal(SIGHUP,OnBreak);signal(SIGINT,OnBreak);
  signal(SIGQUIT,OnBreak);signal(SIGTERM,OnBreak);

  /* Establish timer signal if needed */
  if(SyncFreq) TimerSignal(SyncFreq,OnTimer);

  return(1);
}

/** TrashMachine() *******************************************/
/** Deallocate all resources taken by InitMachine().        **/
/*************************************************************/
void TrashMachine(void)
{
  unsigned long C;
  int J;

  if(Verbose) printf("Shutting down...\n");

  if(Dsp)
  {
    /* Free SCREEN8 palette */
    for(J=0;J<64;J++)
    {
      C=BPal[(J&0x03)|((J&0x0C)<<1)|((J&0x30)<<2)];
      XFreeColors(Dsp,DefaultCMap,&C,1,0);
    }

    X11FreePort(&Prt);
    XAutoRepeatOn(Dsp);
    XCloseDisplay(Dsp);
  }

#ifdef SOUND
  StopSound();TrashSound();
#endif SOUND
}

/** PutImage() ***********************************************/
/** Put an image on the screen.                             **/
/*************************************************************/
void PutImage(void)
{
  /* Wait if needed */
  if(SyncFreq) while(!TimerReady&&!ExitNow);
  /* Copy image */
  X11DrawPort(&Prt,0,0,WIDTH,HEIGHT);
  /* Reset timer flag */
  TimerReady=0;
}

/** Keyboard() ***********************************************/
/** Check for keyboard events, parse them, and modify MSX   **/
/** keyboard matrix.                                        **/
/*************************************************************/
void Keyboard(void)
{
  static byte Control=0;
  XEvent E;
  int J,I;

#ifdef SOUND
  FlushSound();  /* Flush sound stream on each interrupt */
#endif

  /* Check for keypresses/keyreleases */
  if(XCheckWindowEvent(Dsp,Prt.Wnd,KeyPressMask|KeyReleaseMask,&E))
  {
    J=XLookupKeysym((XKeyEvent *)&E,0);
    if(E.type==KeyPress)
    {
      /* Special keys pressed... */
      switch(J)
      {
#ifdef DEBUG
        case XK_F11:   CPU.Trace=!CPU.Trace;break;
#endif

        case XK_F7:
        case XK_F8:
          if(Control)
          {
            I=(J==XK_F8)? 1:0;
            if(Disks[I][0])
            {
              CurDisk[I]=(CurDisk[I]+1)%MAXDISK;
              if(!Disks[I][CurDisk[I]]) CurDisk[I]=0;
              ChangeDisk(I,Disks[I][CurDisk[I]]);
            }
          }
          break;

        case XK_F10:   
          if(Control) UseFont=UseFont? 0:FontBuf? 1:0;
          break;

        case XK_F12:   ExitNow=1;break;

        case XK_Up:    JoyState|=0x01;break;
        case XK_Down:  JoyState|=0x02;break;
        case XK_Left:  JoyState|=0x04;break;
        case XK_Right: JoyState|=0x08;break;
        case XK_Alt_L: JoyState|=0x10;break;
        case XK_Control_L: case XK_Control_R:
          if(J==XK_Control_L) JoyState|=0x20;
          Control=1;
          break;
      }

      /* Key pressed: reset bit in KeyMap */
      J&=0xFF;KeyMap[Keys[J][0]]&=~Keys[J][1];
    }
    else
    {
      /* Special keys released... */
      switch(J)
      {
        case XK_Up:    JoyState&=0xFE;break;
        case XK_Down:  JoyState&=0xFD;break;
        case XK_Left:  JoyState&=0xFB;break;
        case XK_Right: JoyState&=0xF7;break;
        case XK_Alt_L: JoyState&=0xEF;break;
        case XK_Control_L: case XK_Control_R:
          if(J==XK_Control_L) JoyState&=0xDF;
          Control=0;
          break;
      }

      /* Key released: set bit in KeyMap */
      J&=0xFF;KeyMap[Keys[J][0]]|=Keys[J][1];
    }
  }

  /* If window has been resized, change the viewport */
  for(E.type=0;XCheckWindowEvent(Dsp,Prt.Wnd,StructureNotifyMask,&E););
  if(E.type==ConfigureNotify)
    X11ZoomPort(&Prt,E.xconfigure.width,E.xconfigure.height);

  /* If saving CPU and focus is out, sleep */
  for(E.type=0;XCheckWindowEvent(Dsp,Prt.Wnd,FocusChangeMask,&E););
  if(SaveCPU&&(E.type==FocusOut))
  {
    XAutoRepeatOn(Dsp);
#ifdef SOUND
    StopSound();
#endif
  
    do
      while(!XCheckWindowEvent(Dsp,Prt.Wnd,FocusChangeMask,&E)&&!ExitNow)
        sleep(1);
    while((E.type!=FocusIn)&&!ExitNow);

    XAutoRepeatOff(Dsp);
#ifdef SOUND
    ResumeSound();
#endif
  }
}

/** Joystick() ***********************************************/
/** Query position of a joystick connected to port N.       **/
/** Returns 0.0.F2.F1.R.L.D.U.                              **/
/************************************ TO BE WRITTEN BY USER **/
byte Joystick(register byte N) { return(JoyState); }

/** Mouse() **************************************************/
/** Query coordinates of a mouse connected to port N.       **/
/** Returns F2.F1.Y.Y.Y.Y.Y.Y.Y.Y.X.X.X.X.X.X.X.X.          **/
/************************************ TO BE WRITTEN BY USER **/
int Mouse(register byte N)
{
  Window Dummy;
  int X,Y,J,Buttons;

  /* Get mouse coordinates */
  if(!XQueryPointer(Dsp,Prt.Wnd,&Dummy,&Dummy,&J,&J,&X,&Y,&Buttons))
    X=Y=Buttons=0;
  else
  {
    X=X*Prt.W/Prt.WW;
    X-=(Prt.W-256)/2;
    X=X<0? 0:(X>255? 255:X);
    Y=Y*Prt.H/Prt.WH;
    Y-=8;
    Y=Y<0? 0:(Y>211? 211:Y);
    Buttons=(Buttons&Button1Mask? 0x10000:0x00000)
           |(Buttons&Button2Mask? 0x20000:0x00000)
           |(Buttons&Button3Mask? 0x20000:0x00000);
  }

  return((Y<<8)|X|Buttons);
}

/** SetColor() ***********************************************/
/** Set color N (0..15) to R,G,B.                           **/
/*************************************************************/
void SetColor(register byte N,register byte R,register byte G,register byte B)
{
  XPal[0]=XPal0;

  /* Allocate a new color entry */
  XPal[N]=X11GetColor(R,G,B);

  XPal0=XPal[0];
  XPal[0]=(!BGColor||SolidColor0)? XPal0:XPal[BGColor];
}

/** Dummy Functions ******************************************/
/** These dummy functions are called when writes to sound   **/
/** registers occur. Replace them with the actual functions **/
/** generating sound.                                       **/
/*************************************************************/
#ifndef SOUND
void PSGOut(register byte R,register byte V)  { }
void SCCOut(register byte R,register byte V)  { }
#endif
void PPIOut(register byte Toggle)             { }
void OPLLOut(register byte R,register byte V) { }

/** Part of the code common for Unix/X and MSDOS drivers *****/ 
#include "Common.h"

#endif /* UNIX */
